# -*- coding: utf-8 -*-

import numpy as np
from scipy import interpolate
import scipy.stats

from skyllh.core.py import (
    issequenceof,
    float_cast,
    int_cast
)
from skyllh.core.py import module_classname
from skyllh.core.debugging import get_logger
from skyllh.core.signal_generator import SignalGeneratorBase
from skyllh.core.llhratio import LLHRatio
from skyllh.core.dataset import Dataset
from skyllh.core.source_hypothesis import SourceHypoGroupManager
from skyllh.core.storage import DataFieldRecordArray

from skyllh.analyses.i3.publicdata_ps.utils import psi_to_dec_and_ra
from skyllh.analyses.i3.publicdata_ps.smearing_matrix import (
    PDSmearingMatrix
)
from skyllh.analyses.i3.publicdata_ps.aeff import PDAeff


class PDDatasetSignalGenerator(object):
    """This class provides a signal generation method for a point-like source
    seen in the IceCube detector using one dataset of the 10 years public data
    release. It is used by the PDSignalGenerator class in a loop over all the
    datasets that have been added to the analysis.
    """

    def __init__(self, ds, src_dec, effA=None, sm=None, **kwargs):
        """Creates a new instance of the signal generator for generating
        signal events from a specific public data dataset.

        Parameters:
        -----------
        ds : Dataset instance
            Dataset instance for which signal events should get
            generated for.
        src_dec : float
            The declination of the source in radians.
        effA : PDAeff | None
            Representation of the effective area provided by the public data.
        sm : PDSmearingMatrix | None
            Representation of the smearing matrix provided by the public data.
        """
        super().__init__(**kwargs)

        self._logger = get_logger(module_classname(self))

        if sm is None:
            self.smearing_matrix = PDSmearingMatrix(
                pathfilenames=ds.get_abs_pathfilename_list(
                    ds.get_aux_data_definition('smearing_datafile')))
        else:
            self.smearing_matrix = sm

        if effA is None:
            dec_idx = self.smearing_matrix.get_true_dec_idx(src_dec)
            (min_log_true_e,
             max_log_true_e) = \
                self.smearing_matrix.get_true_log_e_range_with_valid_log_e_pdfs(
                    dec_idx)
            kwargs = {
                'src_dec': src_dec,
                'min_log10enu': min_log_true_e,
                'max_log10enu': max_log_true_e
            }
            self.effA = PDAeff(
                pathfilenames=ds.get_abs_pathfilename_list(
                    ds.get_aux_data_definition('eff_area_datafile')),
                **kwargs)

        else:
            self.effA = effA

    def _generate_inv_cdf_spline(self, flux_model, log_e_min,
                                 log_e_max):
        """Sample the true neutrino energy from the power-law
        re-weighted with the detection probability.
        """
        m = (self.effA.log10_enu_bincenters >= log_e_min) & (
            self.effA.log10_enu_bincenters < log_e_max)
        bin_centers = self.effA.log10_enu_bincenters[m]
        low_bin_edges = self.effA._log10_enu_binedges_lower[m]
        high_bin_edges = self.effA._log10_enu_binedges_upper[m]

        # Flux probability P(E_nu | gamma) per bin.
        flux_prob = flux_model.get_integral(
            10**low_bin_edges, 10**high_bin_edges
        ) / flux_model.get_integral(
            10**low_bin_edges[0], 10**high_bin_edges[-1]
        )

        # Do the product and normalize again to a probability per bin.
        product = flux_prob * self.effA.det_prob
        prob_per_bin = product / np.sum(product)

        # The probability per bin cannot be zero, otherwise the cumulative
        # sum would not be increasing monotonically. So we set zero bins to
        # 1000 times smaller than the smallest non-zero bin.
        m = prob_per_bin == 0
        prob_per_bin[m] = np.min(prob_per_bin[np.invert(m)]) / 1000
        to_keep = np.where(prob_per_bin > 1e-15)[0]  # For numerical stability
        prob_per_bin = prob_per_bin[to_keep]
        prob_per_bin /= np.sum(prob_per_bin)

        # Compute the cumulative distribution CDF.
        cum_per_bin = [np.sum(prob_per_bin[:i])
                       for i in range(prob_per_bin.size+1)]
        if np.any(np.diff(cum_per_bin) == 0):
            raise ValueError(
                'The cumulative sum of the true energy probability is not '
                'monotonically increasing! Values of the cumsum are '
                f'{cum_per_bin}.')

        bin_centers = bin_centers[to_keep]
        bin_centers = np.concatenate(([low_bin_edges[0]], bin_centers))

        # Build a spline for the inverse CDF.
        return interpolate.splrep(cum_per_bin, bin_centers, k=1, s=0)

    @staticmethod
    def _eval_spline(x, spl):
        x = np.asarray(x)
        if (x.any() < 0 or x.any() > 1):
            raise ValueError(
                f'{x} is outside of the valid spline range. '
                'The valid range is [0,1].')
        values = interpolate.splev(x, spl, ext=3)
        return values

    def _generate_events(
            self, rss, src_dec, src_ra, dec_idx,
            log_true_e_inv_cdf_spl, n_events):
        """Generates `n_events` signal events for the given source location
        and flux model.

        Note:
            Some values can be NaN in cases where a PDF was not available!

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService to use for drawing random
            numbers.
        src_dec : float
            The declination of the source in radians.
        src_ra : float
            The right-ascention of the source in radians.

        Returns
        -------
        events : numpy record array of size `n_events`
            The numpy record array holding the event data.
            It contains the following data fields:
                - 'isvalid'
                - 'log_true_energy'
                - 'log_energy'
                - 'sin_dec'
            Single values can be NaN in cases where a pdf was not available.
        """

        # Create the output event DataFieldRecordArray.
        out_dtype = [
            ('isvalid', np.bool_),
            ('log_true_energy', np.double),
            ('log_energy', np.double),
            ('dec', np.double),
            ('ra', np.double),
            ('sin_dec', np.double),
            ('ang_err', np.double),
            ('time', int),
            ('azi', np.double),
            ('zen', np.double),
            ('run', int)
        ]

        data = dict(
            [(out_dt[0], np.empty(
                (n_events,),
                dtype=out_dt[1])
              ) for out_dt in out_dtype]
        )

        events = DataFieldRecordArray(data, copy=False)

        sm = self.smearing_matrix

        log_true_e = self._eval_spline(
            rss.random.uniform(size=n_events), log_true_e_inv_cdf_spl)

        events['log_true_energy'] = log_true_e

        log_true_e_idxs = (
            np.digitize(log_true_e, bins=sm.true_e_bin_edges) - 1
        )

        # Sample reconstructed energies given true neutrino energies.
        (log_e_idxs, log_e) = sm.sample_log_e(rss, dec_idx, log_true_e_idxs)
        events['log_energy'] = log_e

        # Sample reconstructed psi values given true neutrino energy and
        # reconstructed energy.
        (psi_idxs, psi) = sm.sample_psi(
            rss, dec_idx, log_true_e_idxs, log_e_idxs)

        # Sample reconstructed ang_err values given true neutrino energy,
        # reconstructed energy, and psi.
        (ang_err_idxs, ang_err) = sm.sample_ang_err(
            rss, dec_idx, log_true_e_idxs, log_e_idxs, psi_idxs)

        isvalid = np.invert(
            np.isnan(log_e) | np.isnan(psi) | np.isnan(ang_err))
        events['isvalid'] = isvalid

        # Convert the psf into a set of (r.a. and dec.). Only use non-nan
        # values.
        (dec, ra) = psi_to_dec_and_ra(rss, src_dec, src_ra, psi[isvalid])
        events['ra'][isvalid] = ra
        events['dec'][isvalid] = dec
        events['sin_dec'][isvalid] = np.sin(dec)

        # Add an angular error. Only use non-nan values.
        events['ang_err'][isvalid] = ang_err[isvalid]

        # Add fields required by the framework
        events['time'] = np.ones(n_events)
        events['azi'] = np.ones(n_events)
        events['zen'] = np.ones(n_events)
        events['run'] = -1 * np.ones(n_events)

        return events

    @staticmethod
    @np.vectorize
    def energy_filter(events, spline, cut_sindec, logger):
        """The energy filter will select all events below `cut_sindec`
        that have an energy smaller than the energy spline at their
        declination.

        Paramters
        ---------
        events : numpy record array
            Numpy record array with the generated signal events.
        energy_cut_splines : scipy.interpolate.UnivariateSpline
            A spline of E(sin_dec) that defines the declination
            dependent energy cut in the IceCube southern sky.
        cut_sindec : float
            The sine of the declination to start applying the energy cut. 
            The cut will be applied from this declination down.
        logger : logging.Logger
            The Logger instance.

        Returns
        energy_filter : (len(events),)-shaped numpy ndarray
            A mask of shape `len(events)` of the events to be cut.
        """
        if cut_sindec is None:
            logger.warn(
                'No `cut_sindec` has been specified. The energy cut will be '
                'applied in [-90, 0] deg.')
            cut_sindec = 0.
        energy_filter = np.logical_and(
            events['sin_dec'] < cut_sindec,
            events['log_energy'] < spline(events['sin_dec']))

        return energy_filter

    def generate_signal_events(
            self, rss, src_dec, src_ra, flux_model, n_events,
            energy_cut_spline=None, cut_sindec=None):
        """Generates ``n_events`` signal events for the given source location
        and flux model.

        Paramters
        ---------
        rss : RandomStateService
        src_dec : float
            Declination coordinate of the injection point.
        src_ra : float
            Right ascension coordinate of the injection point.
        flux_model : FluxModel
            Instance of the `FluxModel` class.
        n_events : int
            Number of signal events to be generated.
        energy_cut_splines : scipy.interpolate.UnivariateSpline
            A spline of E(sin_dec) that defines the declination
            dependent energy cut in the IceCube southern sky.
        cut_sindec : float
            The sine of the declination to start applying the energy cut. 
            The cut will be applied from this declination down.

        Returns
        -------
        events : numpy record array
            The numpy record array holding the event data.
            It contains the following data fields:
                - 'isvalid'
                - 'log_true_energy'
                - 'log_energy'
                - 'dec'
                - 'ra'
                - 'ang_err'
        """
        sm = self.smearing_matrix

        # Find the declination bin index.
        dec_idx = sm.get_true_dec_idx(src_dec)

        # Determine the true energy range for which log_e PDFs are available.
        (min_log_true_e,
         max_log_true_e) = sm.get_true_log_e_range_with_valid_log_e_pdfs(
             dec_idx)
        # Build the spline for the inverse CDF and draw a true neutrino
        # energy from the hypothesis spectrum.
        log_true_e_inv_cdf_spl = self._generate_inv_cdf_spline(
            flux_model, min_log_true_e, max_log_true_e)

        events = None
        n_evt_generated = 0
        while n_evt_generated != n_events:
            n_evt = n_events - n_evt_generated

            events_ = self._generate_events(
                rss, src_dec, src_ra, dec_idx, log_true_e_inv_cdf_spl, n_evt)

            # Cut events that failed to be generated due to missing PDFs.
            # Also cut low energy events if generating in the southern sky.
            events_ = events_[events_['isvalid']]
            if energy_cut_spline is not None:
                to_cut = self.energy_filter(
                    events_, energy_cut_spline, cut_sindec, self._logger)
                events_ = events_[~to_cut]
            if not len(events_) == 0:
                n_evt_generated += len(events_)
                if events is None:
                    events = events_
                else:
                    events.append(events_)

        return events


class PDSignalGenerator(SignalGeneratorBase):
    """This class provides a signal generation method for a point-like source
    seen in the IceCube detector using the 10 years public data release.
    """

    def __init__(self, src_hypo_group_manager, dataset_list, data_list=None,
                 llhratio=None, energy_cut_splines=None, cut_sindec=None):
        """Constructs a new signal generator instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance defining the source hypothesis
            groups.
        dataset_list : list of Dataset instances
            The list of Dataset instances for which signal events should get
            generated for.
        data_list : list of DatasetData instances
            The list of DatasetData instances holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        llhratio : LLHRatio
            The likelihood ratio object contains the datasets signal weights
            needed for distributing the event generation among the different
            datasets.
        energy_cut_splines : list of UnivariateSpline
            A list of splines of E(sin_dec) used to define the declination
            dependent energy cut in the IceCube southern sky.
        cut_sindec : list of float
            The sine of the declination to start applying the energy cut. 
            The cut will be applied from this declination down.
        """
        self.src_hypo_group_manager = src_hypo_group_manager
        self.dataset_list = dataset_list
        self.data_list = data_list
        self.llhratio = llhratio
        self.effA = [None] * len(self._dataset_list)
        self.sm = [None] * len(self._dataset_list)
        self.splines = energy_cut_splines
        self.cut_sindec = cut_sindec

    @property
    def src_hypo_group_manager(self):
        """The SourceHypoGroupManager instance defining the source groups with
        their spectra.
        """
        return self._src_hypo_group_manager

    @src_hypo_group_manager.setter
    def src_hypo_group_manager(self, manager):
        if(not isinstance(manager, SourceHypoGroupManager)):
            raise TypeError('The src_hypo_group_manager property must be an '
                            'instance of SourceHypoGroupManager!')
        self._src_hypo_group_manager = manager

    @property
    def dataset_list(self):
        """The list of Dataset instances for which signal events should get
        generated for.
        """
        return self._dataset_list

    @dataset_list.setter
    def dataset_list(self, datasets):
        if(not issequenceof(datasets, Dataset)):
            raise TypeError('The dataset_list property must be a sequence of '
                            'Dataset instances!')
        self._dataset_list = list(datasets)

    @property
    def llhratio(self):
        """The log-likelihood ratio function for the analysis.
        """
        return self._llhratio

    @llhratio.setter
    def llhratio(self, llhratio):
        if llhratio is not None:
            if(not isinstance(llhratio, LLHRatio)):
                raise TypeError('The llratio property must be an instance of '
                                'LLHRatio!')
        self._llhratio = llhratio

    def generate_signal_events(self, rss, mean, poisson=True):
        shg_list = self._src_hypo_group_manager.src_hypo_group_list
        # Only supports a single source hypothesis group. Raise an error
        # if more than one shg is in the source hypo group manager.
        if len(shg_list) > 1:
            raise RuntimeError(
                'Signal injection for multiple source hypothesis groups is '
                'not supported yet.')

        tot_n_events = 0
        signal_events_dict = {}

        for shg in shg_list:
            # Only supports single point source signal injection. Raise
            # an error if more than one source is in the source hypo group.
            if len(shg.source_list) > 1:
                raise RuntimeError(
                    'Signal injection for multiple sources within a source '
                    'hypothesis group is not supported yet.')
            # This only works with power-laws for now.
            # Each source hypo group can have a different power-law
            gamma = shg.fluxmodel.gamma
            weights, _ = self.llhratio.dataset_signal_weights([mean, gamma])
            for (ds_idx, w) in enumerate(weights):
                w_mean = mean * w
                if(poisson):
                    n_events = rss.random.poisson(
                        float_cast(
                            w_mean,
                            '`mean` must be castable to type of float!'
                        )
                    )
                else:
                    n_events = int_cast(
                        w_mean,
                        '`mean` must be castable to type of int!'
                    )
                tot_n_events += n_events

                events_ = None
                for (shg_src_idx, src) in enumerate(shg.source_list):
                    ds = self._dataset_list[ds_idx]
                    sig_gen = PDDatasetSignalGenerator(
                        ds, src.dec, self.effA[ds_idx], self.sm[ds_idx])
                    if self.effA[ds_idx] is None:
                        self.effA[ds_idx] = sig_gen.effA
                    if self.sm[ds_idx] is None:
                        self.sm[ds_idx] = sig_gen.smearing_matrix
                    # ToDo: here n_events should be split according to some
                    # source weight
                    events_ = sig_gen.generate_signal_events(
                        rss,
                        src.dec,
                        src.ra,
                        shg.fluxmodel,
                        n_events,
                        energy_cut_spline=self.splines[ds_idx],
                        cut_sindec=self.cut_sindec[ds_idx]
                    )
                    if events_ is None:
                        continue

                    if shg_src_idx == 0:
                        signal_events_dict[ds_idx] = events_
                    else:
                        signal_events_dict[ds_idx].append(events_)

        return tot_n_events, signal_events_dict


class PDTimeDependentSignalGenerator(PDSignalGenerator):
    """ The time dependent signal generator works so far only for one single
    dataset. For multi datasets one needs to adjust the dataset weights
    accordingly (scaling of the effective area with livetime of the flare in
    the dataset).
    """

    def __init__(self, src_hypo_group_manager, dataset_list, data_list=None,
                 llhratio=None, energy_cut_splines=None, cut_sindec=None,
                 gauss=None, box=None):
        """
        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        dataset_list : list of Dataset instances
            The list of Dataset instances for which signal events should get
            generated for.
        data_list : list of DatasetData instances
            The list of DatasetData instances holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        llhratio : LLHRatio
            The likelihood ratio object contains the datasets signal weights
            needed for distributing the event generation among the different
            datsets.
        energy_cut_splines : list of UnivariateSpline
        cut_sindec : float
        gauss : dict | None
            None or dictionary with {"mu": float, "sigma": float}.
        box : dict | None
            None or dictionary with {"start": float, "end": float}.
        """
        if gauss is None and box is None:
            raise ValueError(
                "Either box or gauss keywords must define the neutrino flare.")
        if gauss is not None and box is not None:
            raise ValueError(
                "Either box or gauss keywords must define the neutrino flare, "
                "cannot use both.")

        super().__init__(src_hypo_group_manager, dataset_list, data_list,
                         llhratio, energy_cut_splines, cut_sindec)
        self.box = box
        self.gauss = gauss

        self.time_pdf = self._get_time_pdf()

    def _get_time_pdf(self):
        """Get the neutrino flare time pdf given parameters.
        Will be used to generate random numbers by calling `rvs()` method.

        Returns
        -------
        time_pdf : instance of scipy.stats.rv_continuous base class
            Has to base scipy.stats.rv_continuous.
        """
        # Make sure flare is in dataset.
        for data_list in self.data_list:
            grl = data_list.grl

            if self.gauss is not None:
                if (self.gauss["mu"] - 4 * self.gauss["sigma"] > grl["stop"][-1]) or (
                        self.gauss["mu"] + 4 * self.gauss["sigma"] < grl["start"][0]):
                    raise ValueError(
                        f"Gaussian {str(self.gauss)} flare is not in dataset.")

            if self.box is not None:
                if (self.box["start"] > grl["stop"][-1]) or (
                        self.box["end"] < grl["start"][0]):
                    raise ValueError(
                        f"Box {str(self.box)} flare is not in dataset.")

        # Create `time_pdf`.
        if self.gauss is not None:
            time_pdf = scipy.stats.norm(self.gauss["mu"], self.gauss["sigma"])
        if self.box is not None:
            time_pdf = scipy.stats.uniform(
                self.box["start"],
                self.box["end"] - self.box["start"]
            )

        return time_pdf

    def set_flare(self, gauss=None, box=None):
        """Set the neutrino flare given parameters.

        Parameters
        ----------
        gauss : dict | None
            None or dictionary with {"mu": float, "sigma": float}.
        box : dict | None
             None or dictionary with {"start": float, "end": float}.
        """
        if gauss is None and box is None:
            raise ValueError(
                "Either box or gauss keywords must define the neutrino flare.")
        if gauss is not None and box is not None:
            raise ValueError(
                "Either box or gauss keywords must define the neutrino flare, "
                "cannot use both.")

        self.box = box
        self.gauss = gauss

        self.time_pdf = self._get_time_pdf()

    def is_in_grl(self, time, grl):
        """Helper function to check if given times are in the grl ontime.

        Parameters
        ----------
        time : 1d ndarray
            Time values.
        grl : ndarray
            Array of the detector good run list.

        Returns
        -------
        is_in_grl : 1d ndarray
            Boolean mask of `time` in grl ontime.
        """
        def f(time, grl):
            return np.any((grl["start"] <= time) & (time <= grl["stop"]))

        # Vectorize `f`, but exclude `grl` argument from vectorization.
        # This is needed to support `time` as an array argument.
        f_v = np.vectorize(f, excluded=[1])
        is_in_grl = f_v(time, grl)

        return is_in_grl

    def generate_signal_events(self, rss, mean, poisson=True):
        """Same as in PDSignalGenerator, but we assign times here. 
        """
        # Call method from the parent class to generate signal events.
        (tot_n_events, signal_events_dict) = super().generate_signal_events(
            rss, mean, poisson=poisson)

        # Assign times for flare. We can also use inverse transform
        # sampling instead of the lazy version implemented here.
        for (ds_idx, events_) in signal_events_dict.items():
            grl = self.data_list[ds_idx].grl

            # Optimized time injection version, based on csky implementation.
            # https://github.com/icecube/csky/blob/7e969639c5ef6dbb42872dac9b761e1e8b0ccbe2/csky/inj.py#L1122
            times = np.array([])
            n_events = len(events_)
            while len(times) < n_events:
                times = np.concatenate(
                    (times, self.time_pdf.rvs(n_events - len(times),
                                              random_state=rss.random))
                )
                # Check if times is in grl.
                is_in_grl_mask = self.is_in_grl(times, grl)
                times = times[is_in_grl_mask]

            events_["time"] = times
        return tot_n_events, signal_events_dict



# -*- coding: utf-8 -*-

"""This file contains the global configuration dictionary, together with some
convenience utility functions to set different configuration settings.
"""

from astropy import units  # type: ignore
import os.path
import sys
from typing import Any, Dict, Iterator, KeysView, ItemsView, ValuesView

# Try to load the yaml package.
YAML_LOADED = True
try:
    import yaml
except ImportError:
    YAML_LOADED = False

from skyllh.core.py import issequenceof


_BASECONFIG = {
    'multiproc': {
        # The number of CPUs to use for functions that allow multi-processing.
        # If this setting is set to an int value in the range [1, N] this
        # setting will be used if a function's local ncpu setting is not
        # specified.
        'ncpu': None
    },
    'debugging': {
        # The default log format.
        'log_format': (
            '%(asctime)s %(processName)s %(name)s %(levelname)s: '
            '%(message)s'),
        # Flag if detailed debug log messages, i.e. trace log messages, should
        # get generated. This is good for debugging but bad for performance.
        'enable_tracing': False
    },
    'project': {
        # The project's working directory.
        'working_directory': '.'
    },
    'repository': {
        # A base path of repository datasets.
        'base_path': None
    },
    # Definition of the internal units to use. These must match with the units
    # from the monto-carlo data files.
    'internal_units': {
        'angle': units.radian,
        'energy': units.GeV,
        'length': units.cm,
        'time': units.s
    },
    'units': {
        'defaults': {
            # Definition of default units used for fluxes.
            'fluxes': {
                'angle': units.radian,
                'energy': units.GeV,
                'length': units.cm,
                'time': units.s
            }
        }
    },
    'dataset': {
        # Define the data field names of the data set's experimental data,
        # that are required by the analysis.
        'analysis_required_exp_field_names': [
            'run', 'ra', 'dec', 'ang_err', 'time', 'log_energy'
        ],
        # Define the data field names of the data set's monte-carlo data,
        # that are required by the analysis.
        'analysis_required_mc_field_names': [
            'true_ra', 'true_dec', 'true_energy', 'mcweight'
        ]
    },
    # Flag if specific calculations in the core module can be cached.
    'caching': {
        'pdf': {
            'MultiDimGridPDF': False
        }
    }
}


class CFGClass(dict):

    """
    This class holds the global config state.

    The class behaves like a dict, delegating all methods of the dict
    interface to the underlying config dictionary.
    """

    # Keep track of whether this class has been instantiated.

    _is_instantiated = False

    def __init__(self, *args, **kwargs) -> None:
        if CFGClass._is_instantiated:
            raise RuntimeError("Can only instantiate CFGClass once")

        super().__init__(*args, **kwargs)
        CFGClass._is_instantiated = True

    def from_yaml(self, yaml_file: str) -> None:
        """
        Update config with yaml file.

        Parameters:
            yaml_file: str
                Path to yaml file.
        """
        if(YAML_LOADED):
            yaml_config = yaml.load(open(yaml_file), Loader=yaml.SafeLoader)
            self.update(yaml_config)
        else:
            raise ImportError(f'Could not import yaml package. Thus can not'
                              f'import config from yaml file {yaml_file}')

    def from_dict(self, user_dict: Dict[Any, Any]) -> None:
        """
        Creates a config from dictionary.

        Parameters:
            user_dict: dict

        """
        self.update(user_dict)


CFG = CFGClass(_BASECONFIG)


def set_internal_units(
        angle_unit=None, energy_unit=None, length_unit=None, time_unit=None):
    """Sets the units used internally to compute quantities. These units must
    match the units used in the monte-carlo files.

    Parameters
    ----------
    angle_unit : instance of astropy.units.UnitBase | None
        The internal unit that should be used for angles.
        If set to ``None``, the unit is not changed.
    energy_unit : instance of astropy.units.UnitBase | None
        The internal unit that should be used for energy.
        If set to ``None``, the unit is not changed.
    length_unit : instance of astropy.units.UnitBase | None
        The internal unit that should be used for length.
        If set to ``None``, the unit is not changed.
    time_unit : instance of astropy.units.UnitBase | None
        The internal unit that should be used for time.
        If set to ``None``, the unit is not changed.
    """
    if(angle_unit is not None):
        if(not isinstance(angle_unit, units.UnitBase)):
            raise TypeError(
                'The angle_unit argument must be an instance of '
                'astropy.units.UnitBase!')
        CFG['internal_units']['angle'] = angle_unit

    if(energy_unit is not None):
        if(not isinstance(energy_unit, units.UnitBase)):
            raise TypeError(
                'The energy_unit argument must be an instance of '
                'astropy.units.UnitBase!')
        CFG['internal_units']['energy'] = energy_unit

    if(length_unit is not None):
        if(not isinstance(length_unit, units.UnitBase)):
            raise TypeError(
                'The length_unit argument must be an instance of '
                'astropy.units.UnitBase!')
        CFG['internal_units']['length'] = length_unit

    if(time_unit is not None):
        if(not isinstance(time_unit, units.UnitBase)):
            raise TypeError(
                'The time_unit argument must be an instance of '
                'astropy.units.UnitBase!')
        CFG['internal_units']['time'] = time_unit


def set_wd(path):
    """Sets the project's working directory configuration variable and adds it
    to the Python path variable.

    Parameters
    ----------
    path : str
        The path of the project's working directory. This can be a path
        relative to the path given by ``os.path.getcwd``, the current
        working directory of the program.

    Returns
    -------
    wd : str
        The project's working directory.
    """
    if(CFG['project']['working_directory'] in sys.path):
        sys.path.remove(CFG['project']['working_directory'])

    wd = os.path.abspath(path)
    CFG['project']['working_directory'] = wd
    sys.path.insert(0, wd)

    return wd


def add_analysis_required_exp_data_field_names(fieldnames):
    """Adds the given data field names to the set of data field names of the
    experimental data that are required by the analysis.

    Parameters
    ----------
    fieldnames : str | sequence of str
        The field name or sequence of field names that should get added for the
        experimental data.
    """
    if(isinstance(fieldnames, str)):
        fieldnames = [fieldnames]
    elif(not issequenceof(fieldnames, str)):
        raise TypeError(
            'The fieldnames argument must be an instance of str '
            'or a sequence of type str instances!')

    CFG['dataset']['analysis_required_exp_field_names'] = list(set(
        CFG['dataset']['analysis_required_exp_field_names'] + fieldnames))


def set_analysis_required_exp_data_field_names(fieldnames):
    """Sets the data field names of the experimental data that are required by
    the analysis.

    Parameters
    ----------
    fieldnames : str | sequence of str
        The field name or sequence of field names for the experimental data.
    """
    if(isinstance(fieldnames, str)):
        fieldnames = [fieldnames]
    elif(not issequenceof(fieldnames, str)):
        raise TypeError(
            'The fieldnames argument must be an instance of str '
            'or a sequence of type str instances!')

    CFG['dataset']['analysis_required_exp_field_names'] = list(set(fieldnames))


def set_analysis_required_mc_data_field_names(fieldnames):
    """Sets the data field names of the monte-carlo data that are required by
    the analysis.

    Parameters
    ----------
    fieldnames : str | sequence of str
        The field name or sequence of field names for the monte-carlo data.
    """
    if(isinstance(fieldnames, str)):
        fieldnames = [fieldnames]
    elif(not issequenceof(fieldnames, str)):
        raise TypeError(
            'The fieldnames argument must be an instance of str '
            'or a sequence of type str instances!')

    CFG['dataset']['analysis_required_mc_field_names'] = list(set(fieldnames))



# -*- coding: utf-8 -*-

import abc
import numpy as np

from astropy import units

from skyllh.core.py import issequenceofsubclass
from skyllh.core.dataset import Dataset, DatasetData
from skyllh.core.livetime import Livetime
from skyllh.physics.source import SourceModel
from skyllh.physics.flux import FluxModel


def get_integrated_livetime_in_days(livetime):
    """Gets the integrated live-time in days from the given livetime argument.

    Parameters
    ----------
    livetime : float | Livetime instance
        The live-time in days as float, or an instance of Livetime.

    Returns
    -------
    livetime_days : float
        The integrated live-time in days.
    """
    livetime_days = livetime
    if(isinstance(livetime, Livetime)):
        livetime_days = livetime.livetime
    return livetime_days


class DetSigYield(object, metaclass=abc.ABCMeta):
    """This is the abstract base class for a detector signal yield.

    The detector signal yield, Y_s(x_s,p_s), is defined as the expected mean
    number of signal events detected by the detector from a given source at
    position x_s with flux fit parameters p_s.

    To construct a detector signal yield object, four ingredients are
    needed: the dataset holding the monte-carlo data events, a signal flux
    model, the live time, and an implementation method that knows howto contruct
    the actual detector yield in an efficient way.
    In general, the implementation method depends on the detector, the source,
    the flux model with its flux model's signal parameters, and the dataset.
    Hence, for a given detector, source, flux model, and dataset, an appropriate
    implementation method needs to be chosen.
    """

    def __init__(self, implmethod, dataset, fluxmodel, livetime):
        """Constructs a new detector signal yield object. It takes
        the monte-carlo data events, a flux model of the signal, and the live
        time to compute the detector signal yield.

        Parameters
        ----------
        implmethod : instance of DetSigYieldImplMethod
            The implementation method to use for constructing and receiving
            the detector signal yield. The appropriate method depends on
            the used flux model.
        dataset : Dataset instance
            The Dataset instance holding the monte-carlo event data.
        fluxmodel : FluxModel
            The flux model instance. Must be an instance of FluxModel.
        livetime : float | Livetime
            The live-time in days to use for the detector signal yield.
        """
        super(DetSigYield, self).__init__()

        self.implmethod = implmethod
        self.dataset = dataset
        self.fluxmodel = fluxmodel
        self.livetime = livetime

    @property
    def implmethod(self):
        return self._implmethod
    @implmethod.setter
    def implmethod(self, method):
        if(not isinstance(method, DetSigYieldImplMethod)):
            raise TypeError('The implmethod property must be an instance of '
                'DetSigYieldImplMethod!')
        self._implmethod = method

    @property
    def dataset(self):
        """The Dataset instance, for which this detector signal yield is made
        for.
        """
        return self._dataset
    @dataset.setter
    def dataset(self, ds):
        if(not isinstance(ds, Dataset)):
            raise TypeError('The dataset property must be an instance of '
                'Dataset!')
        self._dataset = ds

    @property
    def fluxmodel(self):
        """The flux model, which should be used to calculate the detector
        signal yield.
        """
        return self._fluxmodel
    @fluxmodel.setter
    def fluxmodel(self, model):
        if(not isinstance(model, FluxModel)):
           raise TypeError('The fluxmodel property must be an instance of '
               'FluxModel!')
        self._fluxmodel = model

    @property
    def livetime(self):
        """The live-time in days.
        """
        return self._livetime
    @livetime.setter
    def livetime(self, lt):
        if(not (isinstance(lt, float) or isinstance(lt, Livetime))):
            raise TypeError('The livetime property must be of type float or '
                'an instance of Livetime!')
        self._livetime = lt

    @property
    def n_fitparams(self):
        """(read-only) The number of fit parameters this detector signal yield
        depends on.
        """
        return self._implmethod.n_fitparams

    @property
    def fitparam_names(self):
        """(read-only) The list of fit parameter names this detector signal
        yield depends on.
        """
        return self._implmethod.fitparam_names

    def source_to_array(self, source):
        """Converts the (sequence of) source(s) into a numpy record array needed
        for the __call__ method. This convertion is intrinsic to the
        implementation method.

        Parameters
        ----------
        source : SourceModel | sequence of SourceModel
            The source model containing the spatial information of the source.

        Returns
        -------
        arr : numpy record ndarray
            The generated numpy record ndarray holding the spatial information
            for each source.
        """
        return self._implmethod.source_to_array(source)

    @abc.abstractmethod
    def __call__(self, src, src_flux_params):
        """Abstract method to retrieve the detector signal yield for the given
        sources and source flux parameters.

        Parameters
        ----------
        src : numpy record ndarray
            The numpy record array containing the spatial information of the
            signal sources. The required fields of this record array are
            implementation method dependent. In the most generic case for a
            point-like source, it must contain the following three fields:
            ra, dec, time.
        src_flux_params : numpy record ndarray
            The numpy record ndarray containing the flux parameters of the
            sources. The flux parameters can be different for the different
            sources.

        Returns
        -------
        detsigyield : (N_sources,)-shaped 1D ndarray of float
            The array with the mean number of signal in the detector for each
            given source.
        grads : None | (N_sources,N_fitparams)-shaped 2D ndarray
            The gradient of the detector signal yield w.r.t. each fit
            parameter for each source. If the detector signal yield depends
            on no fit parameter, None is returned.
        """
        pass


class DetSigYieldImplMethod(object, metaclass=abc.ABCMeta):
    """Abstract base class for an implementation method of a detector signal
    yield. Via the ``construct_detsigyield`` method it creates a DetSigYield
    instance holding the internal objects to calculate the detector signal
    yield.
    """

    def __init__(self, **kwargs):
        super(DetSigYieldImplMethod, self).__init__(**kwargs)

        self.supported_sourcemodels = ()
        self.supported_fluxmodels = ()

    @property
    def supported_sourcemodels(self):
        """The tuple with the SourceModel classes, which are supported by this
        detector signal yield implementation method.
        """
        return self._supported_sourcemodels
    @supported_sourcemodels.setter
    def supported_sourcemodels(self, models):
        if(not isinstance(models, tuple)):
            raise TypeError('The supported_sourcemodels property must be of '
                'type tuple!')
        if(not issequenceofsubclass(models, SourceModel)):
            raise TypeError('The supported_sourcemodels property must be a '
                'sequence of SourceModel classes!')
        self._supported_sourcemodels = models

    @property
    def supported_fluxmodels(self):
        """The tuple with the FluxModel classes, which are supported by this
        detector signal yield implementation method.
        """
        return self._supported_fluxmodels
    @supported_fluxmodels.setter
    def supported_fluxmodels(self, models):
        if(not isinstance(models, tuple)):
            raise TypeError('The supported_fluxmodels property must be of '
                'type tuple!')
        if(not issequenceofsubclass(models, FluxModel)):
            raise TypeError('The supported_fluxmodels property must be a '
                'sequence of FluxModel instances!')
        self._supported_fluxmodels = models

    @property
    def n_signal_fitparams(self):
        """(read-only) The number of signal fit parameters the detector signal
        yield depends on.
        """
        return len(self._get_signal_fitparam_names())

    @property
    def signal_fitparam_names(self):
        """(read-only) The list of fit parameter names the detector signal
        yield depends on. An empty list indicates that it does not depend
        on any fit parameter.
        """
        return self._get_signal_fitparam_names()

    def _get_signal_fitparam_names(self):
        """This method must be re-implemented by the derived class and needs to
        return the list of fit parameter names, this detector signal yield
        is a function of. If it returns an empty list, the detector signal
        yield is independent of any fit parameters.

        Returns
        -------
        list of str
            The list of the fit parameter names, this detector signal yield
            is a function of. By default this method returns an empty list
            indicating that the detector signal yield depends on no fit
            parameter.
        """
        return []

    def supports_sourcemodel(self, sourcemodel):
        """Checks if the given source model is supported by this detected signal
        yield implementation method.
        """
        for ssm in self._supported_sourcemodels:
            if(isinstance(sourcemodel, ssm)):
                return True
        return False

    def supports_fluxmodel(self, fluxmodel):
        """Checks if the given flux model is supported by this detector signal
        yield implementation method.
        """
        for sfm in self._supported_fluxmodels:
            if(isinstance(fluxmodel, sfm)):
                return True
        return False

    @abc.abstractmethod
    def construct_detsigyield(self, dataset, data, fluxmodel, livetime):
        """Abstract method to construct the DetSigYield instance.
        This method must be called by the derived class method implementation
        to ensure the compatibility check of the given flux model with the
        supported flux models.

        Parameters
        ----------
        dataset : Dataset
            The Dataset instance holding possible dataset specific settings.
        data : DatasetData
            The DatasetData instance holding the monte-carlo event data.
        fluxmodel : FluxModel
            The flux model instance. Must be an instance of FluxModel.
        livetime : float | Livetime
            The live-time in days to use for the detector signal yield.

        Returns
        -------
        detsigyield : DetSigYield instance
            An instance derived from DetSigYield.
        """
        if(not isinstance(dataset, Dataset)):
            raise TypeError('The dataset argument must be an instance of '
                'Dataset!')
        if(not isinstance(data, DatasetData)):
            raise TypeError('The data argument must be an instance of '
                'DatasetData!')
        if(not self.supports_fluxmodel(fluxmodel)):
            raise TypeError('The DetSigYieldImplMethod "%s" does not support '
                'the flux model "%s"!'%(
                    self.__class__.__name__,
                    fluxmodel.__class__.__name__))
        if((not isinstance(livetime, float)) and
           (not isinstance(livetime, Livetime))):
            raise TypeError('The livetime argument must be an instance of '
                'float or Livetime!')

    @abc.abstractmethod
    def source_to_array(self, source):
        """This method is supposed to convert a (list of) source model(s) into
        a numpy record array that is understood by the implementation method.
        This is for efficiency reasons only. This way the user code can
        pre-convert the list of sources into a numpy record array and cache the
        array.
        The fields of the array are detector signal yield implementation
        dependent, i.e. what kind of sources: point-like source, or extended
        source for instance. Because the sources usually don't change their
        position in the sky, this has to be done only once.

        Parameters
        ----------
        source : SourceModel | sequence of SourceModel
            The source model containing the spatial information of the source.

        Returns
        -------
        arr : numpy record ndarray
            The generated numpy record ndarray holding the spatial information
            for each source.
        """
        pass



# -*- coding: utf-8 -*-

import logging
import os.path
import sys

from skyllh.core.config import CFG


def enable_tracing():
    """Enables the tracing log debug messages of SkyLLH.
    """
    CFG['debugging']['enable_tracing'] = True


def disable_tracing():
    """Disables the tracing log debug messages of SkyLLH.
    """
    CFG['debugging']['enable_tracing'] = False


def is_tracing_enabled():
    """Returns True, if tracing is enabled, False otherwise.
    """
    return CFG['debugging']['enable_tracing']


def get_logger(name):
    """Retrieves the logger with the given name from the Python logging system.

    Parameters
    ----------
    name : str
        The name of the logger.
        Logger hierarchy is defined using dots as separators.

    Returns
    -------
    logger : logging.Logger
        The Logger instance.
    """
    logger = logging.getLogger(name)
    return logger


def setup_logger(name, log_level):
    """Initializes logger with a given name and a log level.

    Parameters
    ----------
    name : str
        Logger name. Loggers hierarchy is defined using dots as separators.
    log_level : int
        The log level.  The ``logging`` module predefines levels, e.g.
        ``logging.DEBUG``.
    """
    logger = logging.getLogger(name)
    logger.setLevel(log_level)


def setup_console_handler(name, log_level=None, log_format=None, stream=None):
    """Initializes `StreamHandler` for a logger with a given name and sets its
    handling level.

    Parameters
    ----------
    name : str
        Logger name. Loggers hierarchy is defined using dots as separators.
    log_level : int | None
        The log level. The ``logging`` module predefines levels, e.g.
        ``logging.DEBUG``.
        If set to None, the log level of the logger will be used.
    log_format : str | None
        The format of log records in the final output.
        If set to `None`, the log format is taken from the configuration.
    stream : data stream | None
        The stream that the handler should use. Default stream is `sys.stderr`.
    """
    logger = logging.getLogger(name)

    if(log_level is None):
        log_level = logger.level

    if(log_format is None):
        log_format = CFG['debugging']['log_format']

    if(stream is None):
        stream = sys.stderr

    # Create and add `StreamHandler` to the logger.
    sh = logging.StreamHandler(stream=stream)
    sh.setLevel(log_level)
    sh.setFormatter(logging.Formatter(log_format))
    logger.addHandler(sh)


def setup_file_handler(
        name, filename, log_level=None, path=None, log_format=None, mode='a'):
    """Initializes `FileHandler` for a logger with a given name and sets its
    handling level.

    Parameters
    ----------
    name : str
        Logger name. Loggers hierarchy is defined using dots as separators.
    log_level : int | None
        The log level. There are predefined levels, e.g. ``logging.DEBUG``.
        If set to None, the log level of the logger will be used.
    filename : str
        The filename of the specified file which is opened and used as the
        stream for logging.
    path : str | None
        The path under which the log file should be stored.
        If set to `None`, the project's working directory will be used.
    log_format : str | None
        The format of log records in the final output.
        If set to `None`, the log format is taken from the configuration.
    mode : str
        File opening mode. Default is 'a' for appending.
    """
    logger = logging.getLogger(name)

    if(log_level is None):
        log_level = logger.level

    if(path is None):
        path = CFG['project']['working_directory']

    if(log_format is None):
        log_format = CFG['debugging']['log_format']

    pathfilename = os.path.join(path, filename)

    # Create and add `FileHandler` to the logger.
    fh = logging.FileHandler(pathfilename, mode=mode)
    fh.setLevel(log_level)
    fh.setFormatter(logging.Formatter(log_format))
    logger.addHandler(fh)


class QueueHandler(logging.Handler):
    """
    This handler sends events to a queue. Typically, it would be used together
    with a multiprocessing Queue to centralise logging to file in one process
    (in a multi-process application), so as to avoid file write contention
    between processes.

    This code is new in Python 3.2, but this class can be copy pasted into
    user code for use with earlier Python versions.
    """
    def __init__(self, queue):
        """
        Initialise an instance, using the passed queue.
        """
        logging.Handler.__init__(self)
        self.queue = queue

    def enqueue(self, record):
        """
        Enqueue a record.

        The base implementation uses put_nowait. You may want to override
        this method if you want to use blocking, timeouts or custom queue
        implementations.
        """
        self.queue.put_nowait(record)

    def prepare(self, record):
        """
        Prepares a record for queuing. The object returned by this method is
        enqueued.

        The base implementation formats the record to merge the message
        and arguments, and removes unpickleable items from the record
        in-place.

        You might want to override this method if you want to convert
        the record to a dict or JSON string, or send a modified copy
        of the record while leaving the original intact.
        """
        # The format operation gets traceback text into record.exc_text
        # (if there's exception data), and also puts the message into
        # record.message. We can then use this to replace the original
        # msg + args, as these might be unpickleable. We also zap the
        # exc_info attribute, as it's no longer needed and, if not None,
        # will typically not be pickleable.
        self.format(record)
        record.msg = record.message
        record.args = None
        record.exc_info = None
        return record

    def emit(self, record):
        """
        Emit a record.

        Writes the LogRecord to the queue, preparing it for pickling first.
        """
        try:
            self.enqueue(self.prepare(record))
        except Exception:
            self.handleError(record)



# -*- coding: utf-8 -*-

"""The time_integrated_ps analysis is a multi-dataset time-integrated single source
analysis with a two-component likelihood function using a spacial and an energy
event PDF.
"""

import argparse
import logging
import numpy as np

from skyllh.core.progressbar import ProgressBar

# Classes to define the source hypothesis.
from skyllh.physics.source import PointLikeSource
from skyllh.physics.flux import PowerLawFlux
from skyllh.core.source_hypo_group import SourceHypoGroup
from skyllh.core.source_hypothesis import SourceHypoGroupManager

# Classes to define the fit parameters.
from skyllh.core.parameters import (
    SingleSourceFitParameterMapper,
    FitParameter
)

# Classes for the minimizer.
from skyllh.core.minimizer import Minimizer, LBFGSMinimizerImpl
from skyllh.core.minimizers.iminuit import IMinuitMinimizerImpl

# Classes for utility functionality.
from skyllh.core.config import CFG
from skyllh.core.random import RandomStateService
from skyllh.core.optimize import SpatialBoxEventSelectionMethod
from skyllh.core.smoothing import BlockSmoothingFilter
from skyllh.core.timing import TimeLord
from skyllh.core.trialdata import TrialDataManager

# Classes for defining the analysis.
from skyllh.core.test_statistic import TestStatisticWilks
from skyllh.core.analysis import (
    TimeIntegratedMultiDatasetSingleSourceAnalysis as Analysis
)

# Classes to define the background generation.
from skyllh.core.scrambling import DataScrambler, UniformRAScramblingMethod
from skyllh.i3.background_generation import FixedScrambledExpDataI3BkgGenMethod

# Classes to define the signal and background PDFs.
from skyllh.core.signalpdf import RayleighPSFPointSourceSignalSpatialPDF
from skyllh.i3.backgroundpdf import (
    DataBackgroundI3SpatialPDF
)

# Classes to define the spatial and energy PDF ratios.
from skyllh.core.pdfratio import SpatialSigOverBkgPDFRatio

# Analysis utilities.
from skyllh.core.analysis_utils import (
    pointlikesource_to_data_field_array
)

# Logging setup utilities.
from skyllh.core.debugging import (
    setup_logger,
    setup_console_handler,
    setup_file_handler
)

# Pre-defined public IceCube data samples.
from skyllh.datasets.i3 import data_samples

# Analysis specific classes for working with the public data.
from skyllh.analyses.i3.publicdata_ps.signal_generator import PDSignalGenerator

from skyllh.analyses.i3.publicdata_ps.detsigyield import (
    PublicDataPowerLawFluxPointLikeSourceI3DetSigYieldImplMethod
)
from skyllh.analyses.i3.publicdata_ps.signalpdf import (
    PDSignalEnergyPDFSet
)
from skyllh.analyses.i3.publicdata_ps.pdfratio import (
    PDPDFRatio
)
from skyllh.analyses.i3.publicdata_ps.backgroundpdf import (
    PDDataBackgroundI3EnergyPDF
)
from skyllh.analyses.i3.publicdata_ps.utils import create_energy_cut_spline


def psi_func(tdm, src_hypo_group_manager, fitparams):
    """Function to calculate the opening angle between the source position
    and the event's reconstructed position.
    """
    ra = tdm.get_data('ra')
    dec = tdm.get_data('dec')

    # Make the source position angles two-dimensional so the PDF value
    # can be calculated via numpy broadcasting automatically for several
    # sources. This is useful for stacking analyses.
    src_ra = tdm.get_data('src_array')['ra'][:, np.newaxis]
    src_dec = tdm.get_data('src_array')['dec'][:, np.newaxis]

    delta_dec = np.abs(dec - src_dec)
    delta_ra = np.abs(ra - src_ra)
    x = (
        (np.sin(delta_dec / 2.))**2. + np.cos(dec) *
        np.cos(src_dec) * (np.sin(delta_ra / 2.))**2.
    )

    # Handle possible floating precision errors.
    x[x < 0.] = 0.
    x[x > 1.] = 1.

    psi = (2.0*np.arcsin(np.sqrt(x)))

    # For now we support only a single source, hence return psi[0].
    return psi[0, :]


def TXS_location():
    src_ra = np.radians(77.358)
    src_dec = np.radians(5.693)
    return (src_ra, src_dec)


def create_analysis(
    datasets,
    source,
    refplflux_Phi0=1,
    refplflux_E0=1e3,
    refplflux_gamma=2.0,
    ns_seed=100.0,
    ns_min=0.,
    ns_max=1e3,
    gamma_seed=3.0,
    gamma_min=1.,
    gamma_max=5.,
    kde_smoothing=False,
    minimizer_impl="LBFGS",
    cut_sindec=None,
    spl_smooth=None,
    cap_ratio=False,
    compress_data=False,
    keep_data_fields=None,
    optimize_delta_angle=10,
    tl=None,
    ppbar=None
):
    """Creates the Analysis instance for this particular analysis.

    Parameters:
    -----------
    datasets : list of Dataset instances
        The list of Dataset instances, which should be used in the
        analysis.
    source : PointLikeSource instance
        The PointLikeSource instance defining the point source position.
    refplflux_Phi0 : float
        The flux normalization to use for the reference power law flux model.
    refplflux_E0 : float
        The reference energy to use for the reference power law flux model.
    refplflux_gamma : float
        The spectral index to use for the reference power law flux model.
    ns_seed : float
        Value to seed the minimizer with for the ns fit.
    ns_min : float
        Lower bound for ns fit.
    ns_max : float
        Upper bound for ns fit.
    gamma_seed : float | None
        Value to seed the minimizer with for the gamma fit. If set to None,
        the refplflux_gamma value will be set as gamma_seed.
    gamma_min : float
        Lower bound for gamma fit.
    gamma_max : float
        Upper bound for gamma fit.
    kde_smoothing : bool
        Apply a KDE-based smoothing to the data-driven background pdf.
        Default: False.
    minimizer_impl : str | "LBFGS"
        Minimizer implementation to be used. Supported options are "LBFGS"
        (L-BFG-S minimizer used from the :mod:`scipy.optimize` module), or
        "minuit" (Minuit minimizer used by the :mod:`iminuit` module).
        Default: "LBFGS".
    cut_sindec : list of float | None
        sin(dec) values at which the energy cut in the southern sky should
        start. If None, np.sin(np.radians([-2, 0, -3, 0, 0])) is used.
    spl_smooth : list of float
        Smoothing parameters for the 1D spline for the energy cut. If None,
        [0., 0.005, 0.05, 0.2, 0.3] is used.
    cap_ratio : bool
        If set to True, the energy PDF ratio will be capped to a finite value
        where no background energy PDF information is available. This will
        ensure that an energy PDF ratio is available for high energies where
        no background is available from the experimental data.
        If kde_smoothing is set to True, cap_ratio should be set to False!
        Default is False.
    compress_data : bool
        Flag if the data should get converted from float64 into float32.
    keep_data_fields : list of str | None
        List of additional data field names that should get kept when loading
        the data.
    optimize_delta_angle : float
        The delta angle in degrees for the event selection optimization methods.
    tl : TimeLord instance | None
        The TimeLord instance to use to time the creation of the analysis.
    ppbar : ProgressBar instance | None
        The instance of ProgressBar for the optional parent progress bar.

    Returns
    -------
    analysis : TimeIntegratedMultiDatasetSingleSourceAnalysis
        The Analysis instance for this analysis.
    """

    # Create the minimizer instance.
    if minimizer_impl == "LBFGS":
        minimizer = Minimizer(LBFGSMinimizerImpl())
    elif minimizer_impl == "minuit":
        minimizer = Minimizer(IMinuitMinimizerImpl(ftol=1e-8))
    else:
        raise NameError(
            f"Minimizer implementation `{minimizer_impl}` is not supported "
            "Please use `LBFGS` or `minuit`.")

    # Define the flux model.
    flux_model = PowerLawFlux(
        Phi0=refplflux_Phi0, E0=refplflux_E0, gamma=refplflux_gamma)

    # Define the fit parameter ns.
    fitparam_ns = FitParameter('ns', ns_min, ns_max, ns_seed)

    # Define the gamma fit parameter.
    fitparam_gamma = FitParameter(
        'gamma', valmin=gamma_min, valmax=gamma_max, initial=gamma_seed)

    # Define the detector signal efficiency implementation method for the
    # IceCube detector and this source and flux_model.
    # The sin(dec) binning will be taken by the implementation method
    # automatically from the Dataset instance.
    gamma_grid = fitparam_gamma.as_linear_grid(delta=0.1)
    detsigyield_implmethod = \
        PublicDataPowerLawFluxPointLikeSourceI3DetSigYieldImplMethod(
            gamma_grid)

    # Define the signal generation method.
    #sig_gen_method = PointLikeSourceI3SignalGenerationMethod()
    sig_gen_method = None

    # Create a source hypothesis group manager.
    src_hypo_group_manager = SourceHypoGroupManager(
        SourceHypoGroup(
            source, flux_model, detsigyield_implmethod, sig_gen_method))

    # Create a source fit parameter mapper and define the fit parameters.
    src_fitparam_mapper = SingleSourceFitParameterMapper()
    src_fitparam_mapper.def_fit_parameter(fitparam_gamma)

    # Define the test statistic.
    test_statistic = TestStatisticWilks()

    # Define the data scrambler with its data scrambling method, which is used
    # for background generation.
    data_scrambler = DataScrambler(UniformRAScramblingMethod())

    # Create background generation method.
    bkg_gen_method = FixedScrambledExpDataI3BkgGenMethod(data_scrambler)

    # Create the Analysis instance.
    analysis = Analysis(
        src_hypo_group_manager,
        src_fitparam_mapper,
        fitparam_ns,
        test_statistic,
        bkg_gen_method,
        sig_generator_cls=PDSignalGenerator
    )

    # Define the event selection method for pure optimization purposes.
    # We will use the same method for all datasets.
    event_selection_method = SpatialBoxEventSelectionMethod(
        src_hypo_group_manager, delta_angle=np.deg2rad(optimize_delta_angle))

    # Prepare the spline parameters.
    if cut_sindec is None:
        cut_sindec = np.sin(np.radians([-2, 0, -3, 0, 0]))
    if spl_smooth is None:
        spl_smooth = [0., 0.005, 0.05, 0.2, 0.3]
    if len(spl_smooth) < len(datasets) or len(cut_sindec) < len(datasets):
        raise AssertionError(
            "The length of the spl_smooth and of the cut_sindec must be equal "
            f"to the length of datasets: {len(datasets)}.")

    # Add the data sets to the analysis.
    pbar = ProgressBar(len(datasets), parent=ppbar).start()
    energy_cut_splines = []
    for idx, ds in enumerate(datasets):
        # Load the data of the data set.
        data = ds.load_and_prepare_data(
            keep_fields=keep_data_fields,
            compress=compress_data,
            tl=tl)

        # Create a trial data manager and add the required data fields.
        tdm = TrialDataManager()
        tdm.add_source_data_field('src_array',
                                  pointlikesource_to_data_field_array)
        tdm.add_data_field('psi', psi_func)

        sin_dec_binning = ds.get_binning_definition('sin_dec')
        log_energy_binning = ds.get_binning_definition('log_energy')

        # Create the spatial PDF ratio instance for this dataset.
        spatial_sigpdf = RayleighPSFPointSourceSignalSpatialPDF(
            dec_range=np.arcsin(sin_dec_binning.range))
        spatial_bkgpdf = DataBackgroundI3SpatialPDF(
            data.exp, sin_dec_binning)
        spatial_pdfratio = SpatialSigOverBkgPDFRatio(
            spatial_sigpdf, spatial_bkgpdf)

        # Create the energy PDF ratio instance for this dataset.
        energy_sigpdfset = PDSignalEnergyPDFSet(
            ds=ds,
            src_dec=source.dec,
            flux_model=flux_model,
            fitparam_grid_set=gamma_grid,
            ppbar=ppbar
        )
        smoothing_filter = BlockSmoothingFilter(nbins=1)
        energy_bkgpdf = PDDataBackgroundI3EnergyPDF(
            data.exp, log_energy_binning, sin_dec_binning,
            smoothing_filter, kde_smoothing)

        energy_pdfratio = PDPDFRatio(
            sig_pdf_set=energy_sigpdfset,
            bkg_pdf=energy_bkgpdf,
            cap_ratio=cap_ratio
        )

        pdfratios = [spatial_pdfratio, energy_pdfratio]

        analysis.add_dataset(
            ds, data, pdfratios, tdm, event_selection_method)

        energy_cut_spline = create_energy_cut_spline(
            ds, data.exp, spl_smooth[idx])
        energy_cut_splines.append(energy_cut_spline)

        pbar.increment()
    pbar.finish()

    analysis.llhratio = analysis.construct_llhratio(minimizer, ppbar=ppbar)
    analysis.construct_signal_generator(
        llhratio=analysis.llhratio, energy_cut_splines=energy_cut_splines,
        cut_sindec=cut_sindec)

    return analysis


if(__name__ == '__main__'):
    p = argparse.ArgumentParser(
        description='Calculates TS for a given source location using the '
        '10-year public point source sample.',
        formatter_class=argparse.RawTextHelpFormatter
    )
    p.add_argument(
        '--dec',
        default=23.8,
        type=float,
        help='The source declination in degrees.'
    )
    p.add_argument(
        '--ra',
        default=216.76,
        type=float,
        help='The source right-ascention in degrees.'
    )
    p.add_argument(
        '--gamma-seed',
        default=3,
        type=float,
        help='The seed value of the gamma fit parameter.'
    )
    p.add_argument(
        '--data_base_path',
        default=None,
        type=str,
        help='The base path to the data samples (default=None)'
    )
    p.add_argument(
        '--seed',
        default=1,
        type=int,
        help='The random number generator seed for the likelihood '
             'minimization.'
    )
    p.add_argument(
        '--ncpu',
        default=1,
        type=int,
        help='The number of CPUs to utilize where parallelization is possible.'
    )
    p.add_argument(
        '--cap-ratio',
        action='store_true',
        help='Switch to cap the energy PDF ratio.')
    p.set_defaults(cap_ratio=False)
    args = p.parse_args()

    # Setup `skyllh` package logging.
    # To optimize logging set the logging level to the lowest handling level.
    setup_logger('skyllh', logging.DEBUG)
    log_format = '%(asctime)s %(processName)s %(name)s %(levelname)s: '\
                 '%(message)s'
    setup_console_handler('skyllh', logging.INFO, log_format)
    setup_file_handler('skyllh', 'debug.log',
                       log_level=logging.DEBUG,
                       log_format=log_format)

    CFG['multiproc']['ncpu'] = args.ncpu

    sample_seasons = [
        ('PublicData_10y_ps', 'IC40'),
        ('PublicData_10y_ps', 'IC59'),
        ('PublicData_10y_ps', 'IC79'),
        ('PublicData_10y_ps', 'IC86_I'),
        ('PublicData_10y_ps', 'IC86_II-VII'),
    ]

    datasets = []
    for (sample, season) in sample_seasons:
        # Get the dataset from the correct dataset collection.
        dsc = data_samples[sample].create_dataset_collection(
            args.data_base_path)
        datasets.append(dsc.get_dataset(season))

    # Define a random state service.
    rss = RandomStateService(args.seed)

    # Define the point source.
    source = PointLikeSource(np.deg2rad(args.ra), np.deg2rad(args.dec))
    print('source: ', str(source))

    tl = TimeLord()

    with tl.task_timer('Creating analysis.'):
        ana = create_analysis(
            datasets,
            source,
            cap_ratio=args.cap_ratio,
            gamma_seed=args.gamma_seed,
            tl=tl)

    with tl.task_timer('Unblinding data.'):
        (TS, fitparam_dict, status) = ana.unblind(rss)

    print('TS = %g' % (TS))
    print('ns_fit = %g' % (fitparam_dict['ns']))
    print('gamma_fit = %g' % (fitparam_dict['gamma']))

    print(tl)



# -*- coding: utf-8 -*-

"""The test_statistic module provides the classes for different test statistic
functions.
"""

import abc
import numpy as np


class TestStatistic(object, metaclass=abc.ABCMeta):
    """This is the abstract base class for a test statistic class.
    """

    def __init__(self):
        """Constructs the test-statistic function instance.
        """
        super(TestStatistic, self).__init__()

    @abc.abstractmethod
    def evaluate(self, llhratio, log_lambda, fitparam_values, *args, **kwargs):
        """This method is supposed to evaluate the test-statistic function.

        Parameters
        ----------
        llhratio : LLHRatio instance
            The log-likelihood ratio function, which should be used for the
            test-statistic function.
        log_lambda : float
            The value of the log-likelihood ratio function. Usually, this is its
            maximum.
        fitparam_values : (N_fitparams+1)-shaped 1D numpy ndarray
            The numpy ndarray holding the fit parameter values of the
            log-likelihood ratio function for the given log_lambda value.

        Returns
        -------
        TS : float
            The calculated test-statistic value.
        """
        pass


class TestStatisticWilks(TestStatistic):
    """This class implements the standard Wilks theorem test-statistic function:

        TS = 2 * sign(ns_best) * log( L(fitparam_best) / L(ns = 0) )

    where the sign(ns_best) is negative for ns_best < 0, and positive otherwise.
    """
    def __init__(self):
        """Constructs the test-statistic function instance.
        """
        super(TestStatisticWilks, self).__init__()

    def evaluate(self, llhratio, log_lambda, fitparam_values):
        """Evaluates this test-statistic function.

        Parameters
        ----------
        llhratio : LLHRatio instance
            The log-likelihood ratio function, which should be used for the
            test-statistic function.
        log_lambda : float
            The value of the log-likelihood ratio function. Usually, this is its
            maximum.
        fitparam_values : (N_fitparams+1)-shaped 1D numpy ndarray
            The numpy ndarray holding the fit parameter values of the
            log-likelihood ratio function for the given log_lambda value.
            By definition, the first element is the value of 'ns'.

        Returns
        -------
        TS : float
            The calculated test-statistic value.
        """
        ns = fitparam_values[0]

        # We need to distinguish between ns=0 and ns!=0, because the np.sign(ns)
        # function returns 0 for ns=0, but we want it to be 1 in such cases.
        sgn_ns = np.where(ns == 0, 1., np.sign(ns))

        TS = 2 * sgn_ns * log_lambda

        return TS


class TestStatisticWilksZeroNsTaylor(TestStatistic):
    """Similar to the TestStatisticWilks class, this class implements the
    standard Wilks theorem test-statistic function. But for zero ns values, the
    log-likelihood ratio function is taylored up to second order and the
    resulting apex is used as log_lambda value. Hence, the TS function is
    defined as:

        TS = 2 * sign(ns_best) * log( L(fitparam_best) / L(ns = 0) )

    for ns_best != 0, and

        TS = 2 * a^2 / (4*b)

    for ns_best == 0, with

        a = d/dns ( L(fitparam_best) / L(ns = 0) )

    being the derivative w.r.t. ns of the log-likelihood ratio function, and

        b = d/dns ( a )

    being its second derivative w.r.t. ns.
    """
    def __init__(self):
        """Constructs the test-statistic function instance.
        """
        super(TestStatisticWilksZeroNsTaylor, self).__init__()

    def evaluate(self, llhratio, log_lambda, fitparam_values, grads):
        """Evaluates this test-statistic function.

        Parameters
        ----------
        llhratio : LLHRatio instance
            The log-likelihood ratio function, which should be used for the
            test-statistic function.
        log_lambda : float
            The value of the log-likelihood ratio function. Usually, this is its
            maximum.
        fitparam_values : (N_fitparams+1)-shaped 1D numpy ndarray
            The numpy ndarray holding the fit parameter values of the
            log-likelihood ratio function for the given log_lambda value.
            By definition, the first element is the value of 'ns'.
        grads : (N_fitparams+1)-shaped 1D ndarray
            The ndarray holding the values of the first derivative of the
            log-likelihood ratio function w.r.t. each global fit parameter.
            By definition the first element is the first derivative
            w.r.t. the fit parameter ns.

        Returns
        -------
        TS : float
            The calculated test-statistic value.
        """
        ns = fitparam_values[0]

        if(ns == 0):
            nsgrad = grads[0]
            nsgrad2 = llhratio.calculate_ns_grad2(fitparam_values)

            TS = -2 * nsgrad**2 / (4*nsgrad2)
        else:
            TS = 2 * np.sign(ns) * log_lambda

        return TS



# -*- coding: utf-8 -*-

import numpy as np

from scipy.interpolate import interp1d


def make_spline_1d(x, y, kind='linear', **kwargs):
    """Creates a 1D spline for the function y(x) using
    :class:`scipy.interpolate.interp1d`.

    Parameters
    ----------
    x : array_like
        The x values.
    y : array_like
        The y values.
    kind : str
        The kind of the spline. See the :class:`scipy.interpolate.interp1d`
        documentation for possible values. Default is ``'linear'``.
    **kwargs
        Additional keyword arguments are passed to the :class:`~scipy.interpolate.interp1d` function.

    Returns
    -------
    spline :
        The created 1D spline instance.
    """
    x = np.atleast_1d(x)
    y = np.atleast_1d(y)

    # The interp1d function requires unique x values. So we need to sort x in
    # increasing order and mask out repeating x values.
    xy = np.array(sorted(zip(x, y)), dtype=y.dtype)
    x = xy[:,0]
    unique_x_mask = np.concatenate(([True], np.invert(
        x[1:] <= x[:-1])))
    x = x[unique_x_mask]
    y = xy[:,1][unique_x_mask]

    spline = interp1d(x, y, kind=kind, copy=False, assume_sorted=True, **kwargs)

    return spline
